package vsl;

import gl4java.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.helpers.*;

import vsl.dataflow.*;
import vsl.gui.*;
import vsl.gui.file.*;
import vsl.util.*;

/*
 * Klasse fr das Hauptprogramm und die Benutzungsschnittstelle.
 */
public class VisualEditor extends JFrame implements WindowListener, ActionListener, MenuListener
{

  /*
   * Hauptmenu
   */
  private JMenuBar menuBar;

  /*
   * Menuleiste "File"
   */
  private JMenu fileMenu;
  private JMenuItem exitMenuItem;

  private JMenu newMenu;
  private JMenuItem newDiagramItem;
  private JMenuItem newFragmentProgramItem;
  private JMenuItem newVertexProgramItem;

  private JMenu openMenu;
  private JMenuItem openDiagramItem;
  private JMenuItem openFragmentProgramItem;
  private JMenuItem openVertexProgramItem;

  private JMenuItem saveItem;
  private JMenuItem saveAsItem;
  private JMenuItem saveAllItem;
  private JMenuItem exportCgItem;

  private JMenuItem closeItem;

  private JMenu editMenu;
  private JMenuItem deleteItem;
  private JMenuItem renameItem;

  /*
   * Menuleiste "Preview"
   */
  private JMenu previewMenu;
  private JMenu setTextureMenu;
  private JMenu textureUnitMenu[];
  private int textureCount = 8;
  private JMenuItem[] setCubeMapItem;
  private JMenuItem[] setTextureItem;
  private JMenu setObjectMenu;

  private JComponent northComponent;
  private JPanel contentPanel;

  private JButton previewButton;

  private Workspace workspace;
  private PreviewCanvas previewCanvas;

  private ProgramCreator vertexCreator;
  private ProgramCreator fragmentCreator;

  private File workingDirectory;

  private JFileChooser cgChooser;
  private JFileChooser textureChooser;
  private JFileChooser cubeMapChooser;
  private JFileChooser vertexChooser;
  private JFileChooser fragmentChooser;
  private JFileChooser diagramChooser;

  private ExtensionFileFilter textureFileFilter;
  private ExtensionFileFilter diagramFileFilter;
  private ExtensionFileFilter vertexProgramFileFilter;
  private ExtensionFileFilter fragmentProgramFileFilter;
  private ExtensionFileFilter cgSourceFileFilter;

  private ShaderTransformer transformer;

  /*
   * Erzeugung der Oberflche.
   */
  public VisualEditor(String title)
  {
    super( title );
    this.addWindowListener( this );
    getContentPane().setLayout( new BorderLayout() );
    setSize( 1024, 768 );

    initMenu();

    workspace = new Workspace();
    setContentPane( workspace );

    previewCanvas = workspace.getPreviewFrame().getPreview();

    vertexCreator = new ProgramCreator( this, "New Vertex Program", true,
                                        ProgramCreator.VERTEX_PROGRAM );
    fragmentCreator = new ProgramCreator( this, "New Fragment Program", true,
                                          ProgramCreator.FRAGMENT_PROGRAM );

    workingDirectory = new File( "." );

    textureChooser = new JFileChooser();
    textureFileFilter = new ExtensionFileFilter( "png", "Texture Files" );
    textureChooser.setFileFilter( textureFileFilter );
    textureChooser.setCurrentDirectory( workingDirectory );

    cubeMapChooser = new JFileChooser();
    cubeMapChooser.setFileFilter( textureFileFilter );
    cubeMapChooser.setMultiSelectionEnabled( true );
    cubeMapChooser.setCurrentDirectory( workingDirectory );

    diagramChooser = new JFileChooser();
    diagramFileFilter = new ExtensionFileFilter( "xcp", "Diagram Files" );
    diagramChooser.setFileFilter( diagramFileFilter );
    diagramChooser.setCurrentDirectory( workingDirectory );

    vertexChooser = new JFileChooser();
    vertexProgramFileFilter = new ExtensionFileFilter( "xvp", "Vertex Program Files" );
    vertexChooser.setFileFilter( vertexProgramFileFilter );
    vertexChooser.setCurrentDirectory( workingDirectory );

    fragmentChooser = new JFileChooser();
    fragmentProgramFileFilter = new ExtensionFileFilter( "xfp", "Fragment Program Files" );
    fragmentChooser.setFileFilter( fragmentProgramFileFilter );
    fragmentChooser.setCurrentDirectory( workingDirectory );

    cgChooser = new JFileChooser();
    cgSourceFileFilter = new ExtensionFileFilter( "cg", "Cg Source Files" );
    cgChooser.setFileFilter( cgSourceFileFilter );
    cgChooser.setCurrentDirectory( workingDirectory );

    transformer = new ShaderTransformer( "xml/transform.xsl" );

    show();
  }

  /*
   * Start des Programms.
   */
  public static void main( String[] args ) throws HeadlessException
  {
    VisualEditor myFrame = new VisualEditor("Visual Editor");
  }

  /*
   * Initialisierung der Menuleisten.
   */
  public void initMenu()
  {
    menuBar = new JMenuBar();
    initFileMenu();
    menuBar.add( fileMenu );
    initEditMenu();
    menuBar.add( editMenu );
    initPreviewMenu();
    menuBar.add( previewMenu );
    setJMenuBar( menuBar );
  }

  public void initFileMenu()
  {
    fileMenu = new JMenu( "File" );

    initNewMenu();
    fileMenu.add( newMenu );

    fileMenu.addSeparator();

    initOpenMenu();
    fileMenu.add( openMenu );

    fileMenu.addSeparator();

    saveItem = new JMenuItem( "Save" );
    saveItem.addActionListener( this );
    fileMenu.add( saveItem );

    saveAsItem = new JMenuItem( "Save As" );
    saveAsItem.addActionListener( this );
    fileMenu.add( saveAsItem );

    saveAllItem = new JMenuItem( "Save All" );
    saveAllItem.addActionListener( this );
    fileMenu.add( saveAllItem );

    exportCgItem = new JMenuItem( "Export Cg" );
    exportCgItem.addActionListener( this );
    fileMenu.add( exportCgItem );

    fileMenu.addSeparator();

    closeItem = new JMenuItem( "Close" );
    closeItem.addActionListener( this );
    fileMenu.add( closeItem );

    fileMenu.addSeparator();

    exitMenuItem = new JMenuItem( "Exit" );
    exitMenuItem.addActionListener( this );
    fileMenu.add( exitMenuItem );
  }

  public void initEditMenu()
  {
    editMenu = new JMenu( "Edit" );
    deleteItem = new JMenuItem( "Delete" );
    deleteItem.addActionListener( this );
    renameItem = new JMenuItem( "Rename Diagram" );
    renameItem.addActionListener( this );
    editMenu.add( deleteItem );
    editMenu.add( renameItem );
  }

  public void initNewMenu()
  {
    newMenu = new JMenu( "New" );
    newDiagramItem = new JMenuItem( "Diagram" );
    newDiagramItem.addActionListener( this );
    newMenu.add( newDiagramItem );
    newFragmentProgramItem = new JMenuItem( "Fragment Program" );
    newFragmentProgramItem.addActionListener( this );
    newMenu.add( newFragmentProgramItem );
    newVertexProgramItem = new JMenuItem( "Vertex Program" );
    newVertexProgramItem.addActionListener( this );
    newMenu.add( newVertexProgramItem );
  }

  public void initOpenMenu()
  {
    openMenu = new JMenu( "Open" );
    openDiagramItem = new JMenuItem( "Diagram" );
    openDiagramItem.addActionListener( this );
    openMenu.add( openDiagramItem );
    openFragmentProgramItem = new JMenuItem( "Fragment Program" );
    openFragmentProgramItem.addActionListener( this );
    openMenu.add( openFragmentProgramItem );
    openVertexProgramItem = new JMenuItem( "Vertex Program" );
    openVertexProgramItem.addActionListener( this );
    openMenu.add( openVertexProgramItem );
  }

  public void initPreviewMenu()
  {
    previewMenu = new JMenu( "Preview" );
    menuBar.add( previewMenu );

    setTextureMenu = new JMenu( "Set Texture" );
    previewMenu.add( setTextureMenu );

    textureUnitMenu = new JMenu[textureCount];
    setTextureItem = new JMenuItem[textureCount];
    setCubeMapItem = new JMenuItem[textureCount];
    for ( int i = 0; i < textureCount; i++ )
    {
      textureUnitMenu[i] = new JMenu( "TextureUnit " + Integer.toString( i ) + ": empty" );
      textureUnitMenu[i].setVisible( true );
      setTextureItem[i] = new JMenuItem( "Set Texture" );
      setTextureItem[i].addActionListener( this );
      textureUnitMenu[i].add( setTextureItem[i] );
      setCubeMapItem[i] = new JMenuItem( "Set CubeMap" );
      setCubeMapItem[i].addActionListener( this );
      textureUnitMenu[i].add( setCubeMapItem[i] );

      setTextureMenu.add( textureUnitMenu[i] );
    }

  }

  public void actionPerformed( ActionEvent e )
  {
    Object source = e.getSource();
    fileMenuAction( source );
    newMenuAction( source );
    openMenuAction( source );
    editMenuAction( source );
    textureMenuAction( source );
  }

  public void fileMenuAction( Object source )
  {
    if ( source == saveItem )
    {
      DataFlowCanvas canvas = workspace.getDiagramFrame().getCanvas();
      String filename = canvas.getFilename();
      if ( filename != null )
      {
        transformer.saveFile( canvas.getDocument(), filename );
      }
      else
      {
        String documentTag = canvas.getDocument().getDocumentElement().getTagName();
        if ( documentTag.equals( "Diagram" ) )
        {
          int result = diagramChooser.showSaveDialog( this );
          if ( result == JFileChooser.APPROVE_OPTION )
          {
            File file = diagramChooser.getSelectedFile();
            transformer.saveFile( canvas.getDocument(), file );
          }
        }
        else if ( documentTag.equals( "FragmentProgram" ) )
        {
          int result = fragmentChooser.showSaveDialog( this );
          if ( result == JFileChooser.APPROVE_OPTION )
          {
            File file = fragmentChooser.getSelectedFile();
            transformer.saveFile( canvas.getDocument(), file );
          }
        }
        else if ( documentTag.equals( "VertexProgram" ) )
        {
          int result = vertexChooser.showSaveDialog( this );
          if ( result == JFileChooser.APPROVE_OPTION )
          {
            File file = vertexChooser.getSelectedFile();
            transformer.saveFile( canvas.getDocument(), file );
          }
        }
      }
    }
    if ( source == saveAsItem )
    {
      DataFlowCanvas canvas = workspace.getDiagramFrame().getCanvas();
      String documentTag = canvas.getDocument().getDocumentElement().getTagName();
      if ( documentTag.equals( "Diagram" ) )
      {
        int result = diagramChooser.showSaveDialog( this );
        if ( result == JFileChooser.APPROVE_OPTION )
        {
          File file = diagramChooser.getSelectedFile();
          transformer.saveFile( canvas.getDocument(), file );
        }
      }
      else if ( documentTag.equals( "FragmentProgram" ) )
      {
        int result = fragmentChooser.showSaveDialog( this );
        if ( result == JFileChooser.APPROVE_OPTION )
        {
          File file = fragmentChooser.getSelectedFile();
          transformer.saveFile( canvas.getDocument(), file );
        }
      }
      else if ( documentTag.equals( "VertexProgram" ) )
      {
        int result = vertexChooser.showSaveDialog( this );
        if ( result == JFileChooser.APPROVE_OPTION )
        {
          File file = vertexChooser.getSelectedFile();
          transformer.saveFile( canvas.getDocument(), file );
        }
      }
    }
    if ( source == exportCgItem )
    {
      DataFlowCanvas canvas = workspace.getDiagramFrame().getCanvas();
      String documentTag = canvas.getDocument().getDocumentElement().getTagName();
      if ( documentTag.equals( "FragmentProgram" ) || documentTag.equals( "VertexProgram" ) )
      {
        int result = cgChooser.showSaveDialog( this );
        if ( result == JFileChooser.APPROVE_OPTION )
        {
          File file = cgChooser.getSelectedFile();
          transformer.transformToFile( canvas.getDocument(), file );
        }
      }
    }
    if ( source == closeItem )
    {
      workspace.getDiagramFrame().removeSelected();
    }
    if ( source == exitMenuItem )
    {
      int quit = ( int ) JOptionPane.showConfirmDialog(
          this,
          "Save all before exiting?" );
      switch ( quit )
      {
        case 0:

          // Save
          System.exit( 0 );
          break;
        case 1:
          System.exit( 0 );
          break;
        case 2:
          break;
      }

    }

  }

  public void newMenuAction( Object source )
  {
    if ( source == newDiagramItem )
    {
      String s = ( String ) JOptionPane.showInputDialog(
          this,
          "Diagram Name:",
          "New Diagram",
          JOptionPane.QUESTION_MESSAGE );
      if ( s != null )
      {
        workspace.getDiagramFrame().addDiagram( new DataFlowDiagram( s ) );
      }

    }
    if ( source == newFragmentProgramItem )
    {
      fragmentCreator.show();
      if ( fragmentCreator.getResult() == ProgramCreator.OK )
      {
        fragmentCreator.reset();
        workspace.getDiagramFrame().addFragmentProgram( new FragmentProgram( fragmentCreator.
            getName(),
            fragmentCreator.getInputStruct(), fragmentCreator.getOutputStruct() ) );
      }
    }
    if ( source == newVertexProgramItem )
    {
      vertexCreator.show();
      if ( vertexCreator.getResult() == ProgramCreator.OK )
      {
        vertexCreator.reset();
        workspace.getDiagramFrame().addVertexProgram( new VertexProgram( vertexCreator.getName(),
            vertexCreator.getInputStruct(), vertexCreator.getOutputStruct() ) );
      }
    }
  }

  public void openMenuAction( Object source )
  {
    if ( source == openDiagramItem )
    {
      int result = diagramChooser.showOpenDialog( this );
      if ( result == JFileChooser.APPROVE_OPTION )
      {
        Document document = getDocument( diagramChooser.getSelectedFile() );
        DataFlowDiagram diagram = new DataFlowDiagram( document.getDocumentElement() );
        workspace.getDiagramFrame().addDiagram( diagram );
      }
    }
    if ( source == openFragmentProgramItem )
    {
      int result = fragmentChooser.showOpenDialog( this );
      if ( result == JFileChooser.APPROVE_OPTION )
      {
        Document document = getDocument( fragmentChooser.getSelectedFile() );
        FragmentProgram fprogram = new FragmentProgram( document );
        workspace.getDiagramFrame().addFragmentProgram( fprogram );
      }
    }
    if ( source == openVertexProgramItem )
    {
      int result = vertexChooser.showOpenDialog( this );
      if ( result == JFileChooser.APPROVE_OPTION )
      {
        Document document = getDocument( vertexChooser.getSelectedFile() );
        VertexProgram vprogram = new VertexProgram( document );
        workspace.getDiagramFrame().addVertexProgram( vprogram );
      }
    }
  }

  public void editMenuAction( Object source )
  {
    if ( source == deleteItem )
    {
      workspace.getDiagramFrame().getCanvas().deleteSelected();
    }
    if ( source == renameItem )
    {
      String s = ( String ) JOptionPane.showInputDialog(
          this,
          "New Name:",
          "Rename",
          JOptionPane.QUESTION_MESSAGE );
      if ( s != null )
      {
        workspace.getDiagramFrame().renameSelected( s );
      }
    }

  }

  int[] faces = new int[]
      {
      GLEnum.GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GLEnum.GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
      GLEnum.GL_TEXTURE_CUBE_MAP_POSITIVE_X, GLEnum.GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
      GLEnum.GL_TEXTURE_CUBE_MAP_POSITIVE_Y, GLEnum.GL_TEXTURE_CUBE_MAP_NEGATIVE_Y};

  public void textureMenuAction( Object source )
  {
    for ( int i = 0; i < textureCount; i++ )
    {
      if ( source == setTextureItem[i] )
      {
        int result = textureChooser.showOpenDialog( this );
        if ( result == textureChooser.APPROVE_OPTION )
        {
          File file = textureChooser.getSelectedFile();
          String filename = file.getAbsolutePath();
          workspace.getPreviewFrame().getPreview().setTexture( filename, i );
          textureUnitMenu[i].setText( "TextureUnit " + Integer.toString( i ) + ": " + file.getName() );
        }
      }
      if ( source == setCubeMapItem[i] )
      {
        int result = cubeMapChooser.showOpenDialog( this );
        if ( result == cubeMapChooser.APPROVE_OPTION )
        {
          File[] files = cubeMapChooser.getSelectedFiles();
          if ( files.length == 6 )
          {
            String[] filenames = new String[6];
            for ( int j = 0; j < 6; j++ )
            {
              filenames[j] = files[j].getAbsolutePath();
              workspace.getPreviewFrame().getPreview().setCubeMap( filenames, i );
              textureUnitMenu[i].setText( "TextureUnit " + Integer.toString( i ) + ": " +
                                          "Cube Map" );
            }
          }
        }
      }
    }
  }

  public static Document getDocument( File file )
  {
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setValidating( true );

      DocumentBuilder builder = factory.newDocumentBuilder();
      builder.setErrorHandler( new DefaultHandler() );
      return builder.parse( file );
    }
    catch ( Exception e )
    {
      System.err.println( e );
    }
    return null;
  }

  public void menuCanceled( MenuEvent e )
  {
  }

  public void menuSelected( MenuEvent e )
  {
  }

  public void menuDeselected( MenuEvent e )
  {
  }

  public void windowClosing( WindowEvent e )
  {
    int quit = ( int ) JOptionPane.showConfirmDialog(
        this,
        "Save all before exiting?" );
    switch ( quit )
    {
      case 0:

        // Save
        System.exit( 0 );
        break;
      case 1:
        System.exit( 0 );
        break;
      case 2:
        break;
    }
  }

  public void windowOpened( WindowEvent e )
  {
  }

  public void windowClosed( WindowEvent e )
  {
  }

  public void windowIconified( WindowEvent e )
  {
  }

  public void windowDeiconified( WindowEvent e )
  {
  }

  public void windowActivated( WindowEvent e )
  {
  }

  public void windowDeactivated( WindowEvent e )
  {
  }
}