package vsl.dataflow;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import org.w3c.dom.*;
import vsl.lang.*;

/**
 * Oberflchenelement zur Anzeige und Bearbeitung von Datenflussdiagrammen.
 */
public class DataFlowCanvas extends JLayeredPane implements MouseListener,
    MouseMotionListener, ActionListener, KeyListener
{


  private JButton closeButton;

  private int actualLayer = 0;

  /*
   * Auswhlen und Verschieben
   */
  private boolean draggingEnabled;
  private Point mousePressedCoordinates = null;
  private Point oldSelectedCoordinates = null;
  private DataFlowComponent selectedComponent = null;
  private DataFlowAnchor sourceAnchor = null;
  private Point oldDraggedCoordinates;
  private boolean connectingEnabled = false;

  /*
   * Tooltip Hilfsmittel
   */
  private Timer toolTipTimer;
  private int toolTipDelay;
  private boolean toolTip = false;
  private Point toolTipCoordinates = new Point( 0, 0 );

  /*
   * Da Datenflussdiagramme geschachtelt sein knnen, wird zwischen
   * dem Hauptdiagramm und dem aktuell bearbeiteten Diagramm unterschieden.
   */
  private DataFlowDiagram actualDiagram;

  /*
   * Hauptdiagramm
   */
  private DataFlowDiagram diagram;

  /*
   * Hauptdiagramm, falls es ein Vertex- oder Fragment-Programm
   */
  private DataFlowProgram program;

  /*
   * Erzeugt neuen Canvas zur Bearbeitung eines Vertex- oder Fragment-Programms.
   */
  public DataFlowCanvas( DataFlowProgram program )
  {
    canvasInit();
    this.program = program;
    this.diagram = program;
    add( program, JLayeredPane.DEFAULT_LAYER );
    setPreferredSize( new Dimension( program.getWidth() + 20, program.getHeight() + 20 ) );
    program.setLocation( 10, 10 );
  }

  /*
   * Erzeugt neuen Canvas zur Bearbeitung eines Diagramms.
   */
  public DataFlowCanvas( DataFlowDiagram diagram )
  {
    canvasInit();
    this.program = null;
    this.diagram = diagram;
    add( diagram, JLayeredPane.DEFAULT_LAYER );
    setPreferredSize( new Dimension( diagram.getWidth() + 20, diagram.getHeight() + 20 ) );
    diagram.setLocation( 10, 10 );
  }

  /*
   * allgemeine Initialisierung
   */
  public void canvasInit()
  {
    addKeyListener( this );
    addMouseListener( this );
    addMouseMotionListener( this );
    toolTipDelay = 3000;
    toolTipTimer = new Timer( toolTipDelay, this );
    toolTipTimer.setRepeats( false );
    setLayout( null );
    setBackground( Color.WHITE );
    closeButton = new JButton( "" );
    closeButton.setSize( 10, 10 );
    closeButton.addActionListener( this );
  }

  public void paint( Graphics g )
  {
    super.paint( g );
    if ( selectedComponent == actualDiagram && actualDiagram != null )
    {
      markSelected( g );
    }
  }

  private void markSelected( Graphics g )
  {
    g.setColor( Color.GRAY );
    Rectangle b = selectedComponent.getBounds();
    g.drawRect( b.x - 4, b.y - 4, 2, 2 );
    g.fillRect( b.x - 4, b.y - 4, 2, 2 );
    g.drawRect( b.x + b.width + 2, b.y - 4, 2, 2 );
    g.fillRect( b.x + b.width + 2, b.y - 4, 2, 2 );
    g.drawRect( b.x - 4, b.y + b.height + 2, 2, 2 );
    g.fillRect( b.x - 4, b.y + b.height + 2, 2, 2 );
    g.drawRect( b.x + b.width + 2, b.y + b.height + 2, 2, 2 );
    g.fillRect( b.x + b.width + 2, b.y + b.height + 2, 2, 2 );
  }

  /*
   * Gibt den Dateiname des Diagramms zurck.
   */
  public String getFilename()
  {
    if ( diagram != null )
    {
      return diagram.getFilename();
    }
    else
    {
      return null;
    }
  }

  /*
   * Erzeugt eine DOM-Dokument des Diagramms.
   */
  public Document getDocument()
  {
    if ( program != null )
    {
      return program.getXMLDocument();
    }
    else if ( diagram != null )
    {
      return diagram.getXMLDocument();
    }
    else
    {
      return null;
    }
  }

  /*
   * Gibt das Hauptdiagramm zurck.
   */
  public DataFlowDiagram getDiagram()
  {
    return diagram;
  }

  /*
   * Fgt in das aktuelle Diagramm einen Knoten ein.
   */
  public void addNode( DataFlowNode node )
  {
    if ( actualDiagram == null )
    {
      actualDiagram = diagram;
    }
    actualDiagram.addNode( node );
    node.setLocation( actualDiagram.getWidth() / 2, actualDiagram.getHeight() / 2 );
    actualDiagram.repaint();
  }

  /*
   * Erzeugt einen Konstruktor-Knoten und fgt ihn ein.
   */
  public void addConstructorNode( Constructor constructor )
  {
    addNode( new DataFlowConstructorNode( constructor ) );
  }

  /*
   * Erzeugt einen Konstanten-Knoten und fgt ihn ein.
   */
  public void addConstantNode( DataValue value )
  {
    addNode( new DataFlowConstantNode( value ) );
  }

  /*
   * Erzeugt einen Funktions-Knoten und fgt ihn ein.
   */
  public void addFunctionNode( Function function )
  {
    addNode( new DataFlowFunctionNode( function ) );
  }

  /*
   * Erzeugt einen Operator-Knoten und fgt ihn ein.
   */
  public void addOperatorNode( Operator operator )
  {
    addNode( new DataFlowOperatorNode( operator ) );
  }

  /*
   * Erzeugt einen OpenGL-State-Knoten und fgt ihn ein.
   */
  public void addStateNode( StateVariable state )
  {
    addNode( new DataFlowStateNode( state ) );
  }

  /*
   * Erzeugt einen Kompositions-Knoten und fgt ihn ein.
   */
  public void addDiagramNode( DataFlowDiagram diagram )
  {
    addNode( new DataFlowDiagramNode( diagram ) );
  }

  /*
   * Erzeugt einen Kompositions-Knoten und fgt ihn ein.
   */
  public void addDiagramNode( String name )
  {
    addNode( new DataFlowDiagramNode( new DataFlowDiagram( name ) ) );
  }

  /*
   * Erzeugt einen Import und fgt ihn dem akuellen Diagramm hinzu.
   */
  public void addImport( String name, DataType type )
  {
    if ( actualDiagram != null )
    {
      DataFlowSourceAnchor imp = new DataFlowSourceAnchor( name, type, null );
      actualDiagram.addImport( imp );
    }
  }

  /*
   * Erzeugt einen Import und fgt ihn dem akuellen Diagramm hinzu.
   */
  public void addImport( String name, DataType type, DataValue defaultValue )
  {
    if ( actualDiagram != null )
    {
      DataFlowSourceAnchor imp = new DataFlowSourceAnchor( name, type, defaultValue, null );
      actualDiagram.addImport( imp );
    }
  }

  /*
   * Fgt dem Diagramm einen Import hinzu.
   */
  public void addImport( DataFlowSourceAnchor imp )
  {
    if ( actualDiagram != null )
    {
      actualDiagram.addImport( imp );
    }
  }

  /*
   * Fgt dem Diagramm einen Export hinzu.
   */
  public void addExport( DataFlowDestinationAnchor export )
  {
    if ( actualDiagram != null )
    {
      actualDiagram.addExport( export );
    }
  }

  /*
   * Gibt wahr zurck, falls Diagramm eine Vertex- oder Fragment-Programm ist.
   */
  public boolean isProgram()
  {
    if ( program == null )
    {
      return false;
    }
    return true;
  }

  /*
   * Ausgewhlte Komponente lschen!
   */
  public void deleteSelected()
  {
    if ( selectedComponent != actualDiagram )
    {
      switch ( selectedComponent.getComponentType() )
      {
        case DataFlowComponent.DF_NODE:
          DataFlowNode node = ( DataFlowNode ) selectedComponent;
          actualDiagram.removeNode( node );
          selectedComponent = actualDiagram;
          actualDiagram.setSelected( null );
          repaint();
          break;
        case DataFlowComponent.DF_ANCHOR:
          break;
        case DataFlowComponent.DF_EDGE:
          DataFlowEdge edge = ( DataFlowEdge ) selectedComponent;
          actualDiagram.removeEdge( edge );
          break;
      }
      actualDiagram.setSelected( null );
      selectedComponent = actualDiagram;
    }
    actualDiagram.repaint();
  }

  /*
   * Doppelklick auf ausgewhltem Element.
   */
  public void doubleClickSelected()
  {
    switch ( selectedComponent.getComponentType() )
    {
      case DataFlowComponent.DF_NODE:
        DataFlowNode node = ( DataFlowNode ) selectedComponent;
        switch ( node.getNodeType() )
        {
          case DataFlowNode.DF_DIAGRAM_NODE:
            actualDiagram.setEnabled( false );
            actualDiagram = ( ( DataFlowDiagramNode ) ( selectedComponent ) ).
                getDiagram();
            this.add( actualDiagram, DEFAULT_LAYER );
            setLayer( actualDiagram, ++actualLayer );
            selectedComponent = actualDiagram;
            actualDiagram.add( closeButton );
            closeButton.setLocation( actualDiagram.getWidth() - 30, 20 );

            break;
        }
        break;
    }
  }

  /*
   * Schliee aktuelle Komposition!
   */
  public void closeDiagram()
  {
    if ( actualDiagram != diagram )
    {
      DataFlowDiagramNode node = actualDiagram.getDiagramNode();
      node.update();
      node.getParent().setEnabled( true );
      selectedComponent = ( DataFlowComponent ) node.getParent();
      actualDiagram.remove( closeButton );
      remove( actualDiagram );
      actualDiagram = ( DataFlowDiagram ) node.getParent();
      if ( actualDiagram != diagram )
      {
        actualDiagram.add( closeButton );
        closeButton.setLocation( actualDiagram.getWidth() - 30, 20 );
      }
    }
    repaint();
  }

  public void mouseClicked( MouseEvent e )
  {
    Component clickedDiagram = getComponentAt( e.getPoint() );
    if ( clickedDiagram != this && clickedDiagram != null )
    {
      if ( clickedDiagram.isEnabled() )
      {
        actualDiagram = ( DataFlowDiagram ) clickedDiagram;
        Component clickedComponent = findComponentAt( e.getPoint() );
        if ( clickedComponent != null && clickedComponent != this )
        {
          if ( clickedComponent.isEnabled() )
          {
            if ( ( ( DataFlowComponent ) clickedComponent ).isSelectable() )
            {
              selectedComponent = ( DataFlowComponent ) clickedComponent;
              if ( selectedComponent != actualDiagram )
              {
                actualDiagram.setSelected( selectedComponent );
                if ( e.getClickCount() == 2 )
                {
                  doubleClickSelected();
                }
              }
              else
              {
                actualDiagram.setSelected( null );
              }
            }
          }
        }
        else
        {
          selectedComponent = null;
        }
      }
    }
    else
    {
      if ( actualDiagram != null )
      {
        actualDiagram.setSelected( null );
      }
      actualDiagram = null;
    }
    repaint();
  }

  public void mouseEntered( MouseEvent e )
  {
  }

  public void mouseExited( MouseEvent e )
  {
  }

  public void mousePressed( MouseEvent e )
  {
    Component pressedDiagram = getComponentAt( e.getPoint() );
    if ( pressedDiagram != this && pressedDiagram != null )
    {
      if ( pressedDiagram.isEnabled() )
      {
        actualDiagram = ( DataFlowDiagram ) pressedDiagram;
        Component pressedComponent = findComponentAt( e.getPoint() );
        if ( pressedComponent != null && pressedComponent != this )
        {
          if ( pressedComponent.isEnabled() )
          {
            DataFlowComponent dfComponent = ( DataFlowComponent ) pressedComponent;
            if ( dfComponent.isSelectable() )
            {
              selectedComponent = ( DataFlowComponent ) pressedComponent;
              if ( selectedComponent != actualDiagram )
              {
                actualDiagram.setSelected( selectedComponent );
              }
              if ( selectedComponent.isDraggable() )
              {
                oldSelectedCoordinates = selectedComponent.getLocation();
                draggingEnabled = true;
              }
            }
            if ( dfComponent.getComponentType() == DataFlowComponent.DF_ANCHOR )
            {
              DataFlowAnchor anchor = ( DataFlowAnchor ) dfComponent;
              if ( anchor.getAnchorType() == DataFlowAnchor.DF_SOURCE_ANCHOR )
              {
                sourceAnchor = anchor;
                connectingEnabled = true;
              }
            }
            mousePressedCoordinates = e.getPoint();
          }
        }
        else
        {
          if ( actualDiagram != null )
          {
            actualDiagram.setSelected( null );
          }
          actualDiagram = null;
        }
      }
    }
    repaint();
  }

  public void mouseReleased( MouseEvent e )
  {
    draggingEnabled = false;
    if ( connectingEnabled )
    {
      connectingEnabled = false;
      Component component = findComponentAt( e.getX(), e.getY() );
      if ( component != null )
      {
        DataFlowComponent destComponent = ( DataFlowComponent ) component;
        if ( destComponent.getComponentType() == DataFlowComponent.DF_ANCHOR )
        {
          DataFlowAnchor anchor = ( DataFlowAnchor ) destComponent;
          if ( anchor.getAnchorType() == DataFlowAnchor.DF_DESTINATION_ANCHOR )
          {
            DataFlowSourceAnchor output = ( DataFlowSourceAnchor ) sourceAnchor;
            DataFlowDestinationAnchor input = ( DataFlowDestinationAnchor ) destComponent;
            if ( input.node != output.node || input.node == null )
            {
              actualDiagram.connect( output, input );
            }
          }
        }
        else
        {
          sourceAnchor = null;
        }
        repaint();
      }
    }
  }

  public void mouseDragged( MouseEvent e )
  {
    if ( draggingEnabled )
    {
      Point oldPosition = getLocation();
      Point newPosition = new Point( oldSelectedCoordinates.x + e.getX() -
                                     mousePressedCoordinates.x,
                                     oldSelectedCoordinates.y + e.getY() -
                                     mousePressedCoordinates.y );
      selectedComponent.setLocation( newPosition );
      repaint();
    }
  }

  public void mouseMoved( MouseEvent e )
  {
  }

  public void actionPerformed( ActionEvent e )
  {
    Object source = e.getSource();
    if ( source == closeButton )
    {
      closeDiagram();
    }
  }

  public void keyPressed( KeyEvent e )
  {
    System.out.println( e );
  }

  public void keyReleased( KeyEvent e )
  {
    System.out.println( e );
  }

  public void keyTyped( KeyEvent e )
  {
    System.out.println( e );
    if ( e.getKeyCode() == KeyEvent.VK_DELETE )
    {
      if ( selectedComponent != actualDiagram )
      {
        DataFlowNode node = ( DataFlowNode ) selectedComponent;
        diagram.remove( node );
      }
    }
  }

}
