package vsl.dataflow;

import org.w3c.dom.*;
import vsl.lang.*;
import vsl.dataflow.util.*;
import vsl.util.*;

/*
 * Oberklasse fr Vertex- und Fragment-Programme
 */
public class DataFlowProgram extends DataFlowDiagram
{

  protected Struct inputStruct;
  protected Struct outputStruct;

  protected String inputStructInstanceName = "IN";
  protected String outputStructInstanceName = "OUT";

  protected DataFlowAnchorList outputStructExports = null;
  protected int outputStructExportCount = 0;

  protected DataFlowAnchorList inputStructImports = null;
  protected int inputStructImportCount = 0;

  public DataFlowProgram( Document document )
  {
    super( document.getDocumentElement().getAttribute( "name" ) );
    Element root = document.getDocumentElement();
    this.inputStruct = Struct.getStruct( getInputStructName( root ) );
    this.outputStruct = Struct.getStruct( getOutputStructName( root ) );
    programInit();
    importDiagram( root );
  }

  public String getInputStructName( Element root )
  {
    NodeList nodeList = root.getElementsByTagName( "InputStruct" );
    if ( nodeList.getLength() > 0 )
    {
      Element inputStructXML = ( Element ) ( nodeList.item( 0 ) );
      return inputStructXML.getAttribute( "type" );
    }
    else
    {
      return null;
    }
  }

  public String getOutputStructName( Element root )
  {
    NodeList nodeList = root.getElementsByTagName( "OutputStruct" );
    if ( nodeList.getLength() > 0 )
    {
      Element inputStructXML = ( Element ) ( nodeList.item( 0 ) );
      return inputStructXML.getAttribute( "type" );
    }
    else
    {
      return null;
    }
  }

  public DataFlowProgram( String name, String inputStructString, String outputStructString )
  {
    super( name );
    this.inputStruct = Struct.getStruct( inputStructString );
    this.outputStruct = Struct.getStruct( outputStructString );
    programInit();
  }

  public DataFlowProgram( String name, Struct inputStruct, Struct outputStruct )
  {
    super( name );
    this.name = name;
    this.inputStruct = inputStruct;
    this.outputStruct = outputStruct;
    programInit();
  }

  public void programInit()
  {
    outputStructExports = new DataFlowAnchorList();
    outputStructExportCount = 0;

    inputStructImports = new DataFlowAnchorList();
    inputStructImportCount = 0;

    for ( int i = 0; i < inputStruct.getInputCount(); i++ )
    {

      DataType dataType = DataType.getDataType( inputStruct.getInputType( i ) );
      String inputName = inputStructInstanceName + "." + inputStruct.getInputName( i );

      DataFlowSourceAnchor imp = new DataFlowSourceAnchor( inputName, dataType, null );
      inputStructImports.addAnchor( imp );

      add( imp.getLabel() );
      add( imp );
      inputStructImportCount++;

    }
    updateImportLocations();

    for ( int i = 0; i < outputStruct.getOutputCount(); i++ )
    {

      DataType dataType = DataType.getDataType( outputStruct.getOutputType( i ) );
      String outputName = outputStructInstanceName + "." + outputStruct.getOutputName( i );

      DataFlowDestinationAnchor exp = new DataFlowDestinationAnchor( outputName,
          dataType, null );
      outputStructExports.addAnchor( exp );

      add( exp.getLabel() );
      add( exp );

      outputStructExportCount++;
    }
    updateExportLocations();
  }

  public DataFlowSourceAnchor getImport( String name )
  {
    if ( imports.containsAnchor( name ) )
    {
      return ( DataFlowSourceAnchor ) imports.getAnchor( name );
    }
    else if ( inputStructImports.containsAnchor( name ) )
    {
      return ( DataFlowSourceAnchor ) inputStructImports.getAnchor( name );
    }
    else
    {
      return null;
    }
  }

  public DataFlowDestinationAnchor getExport( String name )
  {
    if ( exports.containsAnchor( name ) )
    {
      return ( DataFlowDestinationAnchor ) exports.getAnchor( name );
    }
    else if ( outputStructExports.containsAnchor( name ) )
    {
      return ( DataFlowDestinationAnchor ) outputStructExports.getAnchor( name );
    }
    else
    {
      return null;
    }
  }


  public String exportProgram()
  {
    Document document = getXMLDocument();
    ShaderTransformer shaderTransformer = new ShaderTransformer( "xml/transform.xsl" );
    return shaderTransformer.transform( document );
  }

  public void updateImportLocations()
  {
    double importDistance = ( double ) getHeight() /
        ( double ) ( imports.getAnchorCount() + inputStructImportCount + 1 );

    for ( int i = 0; i < inputStructImportCount; i++ )
    {
      DataFlowAnchor imp = inputStructImports.getAnchor( i );
      imp.setLocation( 1, ( int ) ( importDistance * ( double ) ( i + 1 ) ) - imp.getHeight() / 2 );
    }
    for ( int i = 0; i < imports.getAnchorCount(); i++ )
    {
      DataFlowAnchor imp = imports.getAnchor( i );
      imp.setLocation( 1,
                       ( int ) ( importDistance * ( double ) ( i + inputStructImportCount + 1 ) ) -
                       imp.getHeight() / 2 );
    }
  }

  public void updateExportLocations()
  {
    double exportDistance = ( double ) getHeight() /
        ( double ) ( exports.getAnchorCount() + outputStructExportCount + 1 );
    for ( int i = 0; i < outputStructExportCount; i++ )
    {
      DataFlowAnchor export = outputStructExports.getAnchor( i );
      export.setLocation( getWidth() - export.getWidth() -1,
                          ( int ) ( exportDistance * ( double ) ( i + 1 ) ) -
                          export.getHeight() / 2 );
    }
    for ( int i = 0; i < exports.getAnchorCount(); i++ )
    {
      DataFlowAnchor export = exports.getAnchor( i );
      export.setLocation( getWidth() - export.getWidth()-1,
                          ( int ) ( exportDistance * ( double ) ( i + outputStructExportCount + 1 ) ) -
                          export.getHeight() / 2 );
    }
  }

  public void recalculateDistances()
  {
    for ( int i = 0; i < outputStructExportCount; i++ )
    {
      DataFlowDestinationAnchor export = ( DataFlowDestinationAnchor ) outputStructExports.
          getAnchor( i );
      if ( export.isConnected() )
      {
        if ( export.getEdge().getSource().hasNode() )
        {
          export.getEdge().getSource().getNode().calculateMaxDistance( 1 );
        }
      }
    }

    for ( int i = 0; i < exports.getAnchorCount(); i++ )
    {
      DataFlowDestinationAnchor export = ( DataFlowDestinationAnchor ) exports.getAnchor( i );
      if ( export.isConnected() )
      {
        if ( export.getEdge().getSource().hasNode() )
        {
          export.getEdge().getSource().getNode().calculateMaxDistance( 1 );
        }
      }
    }
  }

  public void print()
  {
    for ( int i = 0; i < outputStructExportCount; i++ )
    {
      DataFlowDestinationAnchor export = ( DataFlowDestinationAnchor ) outputStructExports.
          getAnchor( i );
      if ( export.isConnected() )
      {
        if ( export.getEdge().getSource().hasNode() )
        {
          export.getEdge().getSource().getNode().print();
        }
      }
    }
    for ( int i = 0; i < exports.getAnchorCount(); i++ )
    {
      DataFlowDestinationAnchor export = ( DataFlowDestinationAnchor ) exports.getAnchor( i );
      if ( export.isConnected() )
      {
        if ( export.getEdge().getSource().hasNode() )
        {
          export.getEdge().getSource().getNode().print();
        }
      }
    }
  }

}